/*
 * This file is a part of Winix
 * and is not publicly distributed
 * 
 * Copyright (c) 2010, Tomasz Sowa
 * All rights reserved.
 * 
 */

#include "tdb.h"
#include "core/log.h"



namespace Ticket
{



Error TDb::GetTicket(long dir_id, Ticket & ticket)
{
	PGresult * r = 0;
	Error status = WINIX_ERR_OK;
	ticket.Clear();
	
	try
	{
		query.Clear();
		query << R("select dir_id, param, value from plugins.ticket "
				   "where ticket.dir_id=") << dir_id << R(";");
		
		r = AssertQuery(query);
		AssertResult(r, PGRES_TUPLES_OK);

		int rows = Rows(r);

		if( rows == 0 )
			throw Error(WINIX_ERR_NO_TICKET);
	
		Ticket::TicketParam par;
		ticket.dir_id = dir_id;

		int cparam  = AssertColumn(r, "param");
		int cvalue  = AssertColumn(r, "value");

		for(int i=0 ; i<rows ; ++i)
		{
			par.param = AssertValueInt(r, i, cparam);
			par.value = AssertValueInt(r, i, cvalue);
			ticket.par_tab.push_back(par);
		}
	}
	catch(const Error & e)
	{
		status = e;
	}
	
	ClearResult(r);

return status;
}






Error TDb::GetTickets(const std::vector<long> & dir_id_tab, std::vector<Ticket> & ticket_tab)
{
	PGresult * r = 0;
	Error status = WINIX_ERR_OK;
	ticket_tab.clear();
	
	if( dir_id_tab.empty() )
		return status;

	try
	{
		query.Clear();
		CreateIdList(dir_id_tab, dir_list);
		query << R("select dir_id, param, value from plugins.ticket "
				   "where ticket.dir_id in ") << R(dir_list) << R(" order by dir_id;");

		// !! moze dodac operator <<(std::vector(long)) do query?

		r = AssertQuery(query);
		AssertResult(r, PGRES_TUPLES_OK);

		int rows = Rows(r);
		Ticket::TicketParam par;

		int cdir_id = AssertColumn(r, "dir_id");
		int cparam  = AssertColumn(r, "param");
		int cvalue  = AssertColumn(r, "value");
		long last_dir_id = -1;

		for(int i=0 ; i<rows ; ++i)
		{
			long dir_id = AssertValueLong(r, i, cdir_id);

			if( i==0 || last_dir_id != dir_id )
			{
				ticket_tab.push_back(Ticket());
				ticket_tab.back().dir_id = dir_id;
				last_dir_id = dir_id;
			}

			par.param = AssertValueInt(r, i, cparam);
			par.value = AssertValueInt(r, i, cvalue);
			ticket_tab.back().par_tab.push_back(par);
		}
	}
	catch(const Error & e)
	{
		status = e;
	}
	
	ClearResult(r);
	dir_list.clear();

return status;
}




Error TDb::AddTicket(const Ticket & ticket)
{
	PGresult * r = 0;
	Error status = WINIX_ERR_OK;
	
	try
	{
		for(size_t i=0 ; i<ticket.par_tab.size() ; ++i)
		{
			query.Clear();
			query << R("insert into plugins.ticket (dir_id, param, value) values (")
				  << ticket.dir_id
				  << ticket.par_tab[i].param
				  << ticket.par_tab[i].value
				  << R(");");
			
			r = AssertQuery(query);
			AssertResult(r, PGRES_COMMAND_OK);
		}
	}
	catch(const Error & e)
	{
		status = e;
	}
	
	ClearResult(r);

return status;
}




Error TDb::RemoveAddTicket(const Ticket & ticket)
{
	Error status = BeginTrans();

	if( status != WINIX_ERR_OK )
		return status;

	status = RemoveTicket(ticket.dir_id);

	if( status == WINIX_ERR_OK )
		status = AddTicket(ticket);

	if( status == WINIX_ERR_OK )
		status = CommitTrans();
	else
		RollbackTrans();

return status;
}





Error TDb::RemoveTicket(long dir_id)
{
	PGresult * r = 0;
	Error status = WINIX_ERR_OK;
	
	try
	{
		query.Clear();
		query << R("delete from plugins.ticket where dir_id=") << dir_id << R(";");

		r = AssertQuery(query);
		AssertResult(r, PGRES_COMMAND_OK);

		long rows = AffectedRows(r);

		if( rows > 0 )
			log << log2 << "Db: deleted " << rows << " rows from plugins.ticket" << logend;	
		
	}
	catch(const Error & e)
	{
		status = e;
	}
	
	ClearResult(r);

return status;
}





} // namespace

